==> cscdata-0.1.6-py3-none-any.whl: cscdata/localdata/stock/context_stk.py <==
# -*- coding:utf-8 -*-
# Author: ranpeng
# Date: 2023/11/13

from cscdata.localdata.context import CoContext
import cscdata.localdata as localdata
from functools import cached_property
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import os

from .utils import QtCalendar, write_json, retry, context_data

today = int(datetime.today().strftime("%Y%m%d"))
yesterday = int((datetime.today() - timedelta(days = 1)).strftime("%Y%m%d"))

class FinContext(CoContext):
    def __init__(self, start_dt=20070101, end_dt= yesterday):
        super().__init__(start_dt, end_dt)

        self.start_dt = start_dt
        self.end_dt = end_dt

        self.callib = None
        self.all_dts = None  # 1990年以来所有的日历日   # v1.4
        self.cn_tds = None  # 1990年以来过去未来全部交易日
        self.ann_dts = None  # [start_dt, end_dt]之间的所有日期日
        self.trade_dts = None  # [start_dt, end_dt]之间的所有交易日
        self.instruments = None  # [start_dt, end_dt]之间的所有上市股票
        self.report_periods = None  # [start_dt, end_dt]外推所有报告期（季度）; 外推3年
        self.report_years = None  # [start_dt, end_dt]外推所有年度报告期

        self.na = 0  # num of 日历日
        self.nd = 0  # num of 交易日
        self.ns = 0  # num of 股票数量
        self.nq = 0  # num of ann_dts

        self.nd_dict = None  # trade_dts反查表
        self.na_dict = None  # ann_dts反查表
        self.ns_dict = None  # instruments反查表
        self.nq_dict = None  # report_periods反查表

        self.ndns_pkeys = ['stockcode', 'trade_dt']
        self.nans_pkeys = ['stockcode', 'ann_dt']
        self.nansnq_pkeys = ['stockcode', 'ann_dt', 'report_period']

        self.ndx_idx = None
        self.univ_ipo = None
        self.nans_ipo = None
        self.nans_idx = None

    @context_data
    def get_univ_ipo_mat(self):
        ''' univ_ipo to ndns_df  上市股票全集'''
        result = self.to_ndns(self.ndns_idx['univ_ipo'], ndns_sorted=False)
        return result
    
    @context_data
    def get_univ_ipo(self):
        '''获取ndns轴， 添加listed'''
        df = self.wind_des.copy()
        df.rename(columns={'list_dt': 'indate', 'delist_dt': 'outdate'}, inplace=True)
        ndns_df = self.get_univ(df[['stockcode', 'indate', 'outdate']], right_close=False)
        result = pd.DataFrame({'univ_ipo': ndns_df.unstack(level=1)})
        return result

    @cached_property
    def wind_cal(self):
        return self.ds('asharecalendar').get_df(index=False)

    @cached_property
    def wind_des(self):
        return self.ds('asharedescription').get_df(index=False)

    def get_instruments(self, start_dt, end_dt):
        ''' 区间内上市的股票 '''
        df = self.wind_des
        return df[(df.delist_dt >= start_dt) & (df.list_dt <= end_dt)].stockcode.values

    def init_from_wind(self):
        # self.db.list_tables()
        df = self.wind_cal
        self.cn_tds = df.loc[df.exch == 'SSE'].sort_values('trade_dt')['trade_dt'].values
        self.callib = QtCalendar(self.wind_cal, all_dts=self.cn_tds)
        self.trade_dts = self.callib.get_trade_dts(self.start_dt, self.end_dt)
        self.instruments = self.get_instruments(self.start_dt, self.end_dt)
        end = self.end_dt + 30000
        rps = pd.date_range(str(self.start_dt), str(end), freq='Q-DEC')  # 外推3年 用于存预期数据
        rys = pd.date_range(str(self.start_dt), str(end), freq='Y')
        self.all_dts = pd.date_range('19900101', str(self.end_dt)).strftime("%Y%m%d").astype(int).values  # v1.4
        self.report_periods = rps.strftime("%Y%m%d").astype(int).values
        self.report_years = rys.strftime("%Y").astype(int).values
        self.ann_dts = pd.date_range(str(self.start_dt), str(self.end_dt)).strftime("%Y%m%d").astype(int).values
        self.nd, self.ns, self.nq, self.na = len(self.trade_dts), len(self.instruments), len(self.report_periods), len(
            self.ann_dts)
        self.nd_dict = pd.Series(np.arange(self.nd), self.trade_dts)
        self.ns_dict = pd.Series(np.arange(self.ns), self.instruments)
        self.nq_dict = pd.Series(np.arange(self.nq), self.report_periods)
        self.na_dict = pd.Series(np.arange(self.na), self.ann_dts)
        self.na_dict1 = pd.Series(np.arange(len(self.all_dts)), self.all_dts)  # v1.4

    @retry(max_attempts=5, delay=1)
    def to_feature(self, ds, feature_name, database = None, des=''):
        """
            将df转换为非结构化数据
        """
        if database is None:
            database = self.user_db_name
        print(f'writing {feature_name}')
        idx_names = ds.index.names
        df = ds.reset_index().rename(columns={0: 'value'})
        df['feature'] = feature_name
        df.drop_duplicates(subset=idx_names, inplace=True)
        self.to_parquet(database + '.feature', df, partition_by=['feature'])

        # json_file =  os.path.join(self.user_save_path, database, 'feature', f'feature={feature_name}', 'des.json' )
        # metadata = {}
        # metadata['primary_keys'] = idx_names
        # metadata['des'] = des
        # metadata['last_update'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        # write_json(metadata, json_file)

    def is_tds(self, dts: np.ndarray, exclude=(0, 99999999)):
        '''是否日期序列均为交易日'''
        exclude_mask = np.isin(dts, exclude)
        return np.isin(dts[~exclude_mask], self.cn_tds).all()

    def cal2td(self, dts):
        ''' 日历日期转换为trade dt '''
        return self.trade_dts[self.dt_loc(dts)]

    def isin_instrument(self, stklist):
        ''' 是否在股票池中 '''
        return np.isin(stklist, self.instruments)

    def pivot_dup(self, df, aggfunc='sum'):
        ''' pivot: with duplicates '''
        df['trade_dt'] = self.cal2td(df['trade_dt'])
        name = df.columns.drop(['stockcode', 'trade_dt'])[0]
        ndns_df = df.pivot_table(index='trade_dt', columns='stockcode', values=name, aggfunc=aggfunc) \
            .reindex(index=self.trade_dts, columns=self.instruments)
        return ndns_df

    def shift_tds(self, dts, n):
        '''把日期变成，前后n个交易日， n不能为0
        n = -1： 向前找最近的一个交易日，
        n = 0： 向后找最近的一个交易日
        如果不在交易日为首日 越界返回0和99999999
        np.searchsorted会找到最近日期靠后的一天的位置
        '''
        dts = dts.values
        tds = self.cn_tds.copy()
        loc = np.searchsorted(tds, dts) + n
        overupper = loc > len(tds) - 1
        belowlower = loc < 0
        normal = ~(overupper | belowlower)

        result = np.ones_like(dts)
        result[normal] = tds[loc[normal]]
        result[overupper] = 99999999
        result[belowlower] = 0
        return result

    def get_univ(self, df, right_close=True):
        ''' df中包含stockcode, indate, outdate; 日期nan的位置填充9999999
        right_close: outdate是否是右闭的
        '''
        assert (np.isin(['stockcode', 'indate', 'outdate'], df.columns).all())
        if not right_close:
            df['outdate'] = self.shift_tds(df['outdate'], -1)
            assert (self.is_tds(df['outdate']))
        else:
            assert (self.is_tds(df['outdate']))
        df.indate = df.indate.clip(lower=self.trade_dts[0])
        df.outdate = df.outdate.clip(upper=self.trade_dts[-1])
        df = df[df['outdate'] >= df['indate']]
        df = df[self.isin_instrument(df['stockcode'])]  # 退市后依旧在

        indate_df = df[['stockcode', 'indate']].rename(columns={'indate': 'trade_dt'})
        indate_df['value'] = 1.0
        indate_mat = self.pivot_dup(indate_df)
        outdate_df = df[['stockcode', 'outdate']].rename(columns={'outdate': 'trade_dt'})
        outdate_df['value'] = 1.0
        outdate_mat = self.pivot_dup(outdate_df)
        indate_mat.fillna(0, inplace=True)
        outdate_mat.fillna(0, inplace=True)

        result = indate_mat.cumsum() - outdate_mat.cumsum()
        result += outdate_mat

        result.replace(0, np.nan, inplace=True)
        result[result > 0] = 1.0
        return result

    def dt_loc(self, dts):
        ''' 获取 dts 在 trade_dts 上的位置，如果找不到取后值位置
        nan的位置变nan
        '''
        loc = np.searchsorted(self.trade_dts, dts)
        return np.clip(loc, a_min=0, a_max=self.nd - 1)

    def stk_loc(self, stks):
        '''返回instrument上门的排序'''
        return self.ns_dict[stks].values

    def get_ind(self, df, on):
        ''' 生成ndns的行业分类要求字段：['ind', 'stockcode', 'indate', 'outdate'];都是右闭 '''
        assert (np.isin(['stockcode', 'indate', 'outdate'], df.columns).all())
        df = df[['stockcode', 'indate', 'outdate', on]].rename(columns={on: 'ind'})

        df.indate = df.indate.clip(lower=self.trade_dts[0])
        df.outdate = df.outdate.clip(upper=self.trade_dts[-1])
        df = df[df['outdate'] >= df['indate']]
        df = df[self.isin_instrument(df['stockcode'])]
        inds = np.unique(df['ind'])
        indmap = pd.Series(np.arange(len(inds)), index=inds)

        df['indate_loc'] = self.dt_loc(df['indate'])
        df['outdate_loc'] = self.dt_loc(df['outdate'])
        df['stk_loc'] = self.stk_loc(df['stockcode'])
        df['weight'] = df['ind'].replace(indmap)

        result = np.empty([self.nd, self.ns]) * np.nan
        for i_row, row in df.iterrows():
            mat_i = np.arange(row['indate_loc'], row['outdate_loc'] + 1)  # 右闭
            mat_j = [int(row['stk_loc'])]
            idx = np.ix_(mat_i, mat_j)
            result[idx] = row['weight']
        ndns_df = pd.DataFrame(result, index=self.trade_dts, columns=self.instruments)
        return ndns_df

    def to_ndns(self, ds, ndns_sorted=True):
        ''' ds: with ndnp_idx 变成ndns_df
        ndns_sorted: False: 代表ndns有序
        '''
        # name = ds.name
        assert (isinstance(ds, pd.Series))
        if ndns_sorted:
            assert (ds.index.names == self.ndns_pkeys)
            v = ds.values.reshape([self.ns, self.nd]).T
            result = pd.DataFrame(v, index=self.trade_dts, columns=self.instruments)
        else:
            result = ds.unstack(level=0).reindex(index=self.trade_dts, columns=self.instruments)
        return result

    def to_nans(self, ds, nans_sorted=True):
        ''' ds: with ndnp_idx 变成ndns_df
        ndns_sorted: True: 代表nans有序
        '''
        # name = ds.name
        assert (isinstance(ds, pd.Series))
        if nans_sorted:
            assert (ds.index.names == self.nans_pkeys)
            v = ds.values.reshape([self.ns, self.na]).T
            result = pd.DataFrame(v, index=self.ann_dts, columns=self.instruments)
        else:
            result = ds.unstack(level=0).reindex(index=self.ann_dts, columns=self.instruments)
        return result

    def run(self,file):
        """run cscdata init first, then use func to init here params"""
        super().run(file)

        self.init_from_wind()
        self.ndns_idx = self.get_univ_ipo()
        
        self.univ_ipo = self.get_univ_ipo_mat()
        self.nans_ipo = self.univ_ipo.reindex(index = self.ann_dts)
        self.nans_idx = pd.DataFrame({'univ_ipo' : self.nans_ipo.unstack(level=1)})
        self.nans_idx.index = self.nans_idx.index.set_names(self.nans_pkeys)
        
        # 必须传入的配置时间轴，为取历史数据使用
        localdata.initialize(self.trade_dts)

        # users func
        self._on("_init")

==> cscdata-0.1.7-py3-none-any.whl: cscdata/localdata/stock/context_stk.py <==
# -*- coding:utf-8 -*-
# Author: ranpeng
# Date: 2023/11/13

from cscdata.localdata.context import CoContext
import cscdata.localdata as localdata
from functools import cached_property
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import os

from .utils import QtCalendar, read_json, write_json, retry, context_data

today = int(datetime.today().strftime("%Y%m%d"))
yesterday = int((datetime.today() - timedelta(days = 1)).strftime("%Y%m%d"))

class FinContext(CoContext):
    def __init__(self, config = None, start_dt=20070101, end_dt= yesterday):
        super().__init__(start_dt, end_dt, config)

        self.start_dt = start_dt
        self.end_dt = end_dt

        self.callib = None
        self.all_dts = None  # 1990年以来所有的日历日   # v1.4
        self.cn_tds = None  # 1990年以来过去未来全部交易日
        self.ann_dts = None  # [start_dt, end_dt]之间的所有日期日
        self.trade_dts = None  # [start_dt, end_dt]之间的所有交易日
        self.instruments = None  # [start_dt, end_dt]之间的所有上市股票
        self.report_periods = None  # [start_dt, end_dt]外推所有报告期（季度）; 外推3年
        self.report_years = None  # [start_dt, end_dt]外推所有年度报告期

        self.na = 0  # num of 日历日
        self.nd = 0  # num of 交易日
        self.ns = 0  # num of 股票数量
        self.nq = 0  # num of ann_dts

        self.nd_dict = None  # trade_dts反查表
        self.na_dict = None  # ann_dts反查表
        self.ns_dict = None  # instruments反查表
        self.nq_dict = None  # report_periods反查表

        self.ndns_pkeys = ['stockcode', 'trade_dt']
        self.nans_pkeys = ['stockcode', 'ann_dt']
        self.nansnq_pkeys = ['stockcode', 'ann_dt', 'report_period']

        self.ndx_idx = None
        self.univ_ipo = None
        self.nans_ipo = None
        self.nans_idx = None

    @context_data
    def get_univ_ipo_mat(self):
        ''' univ_ipo to ndns_df  上市股票全集'''
        result = self.to_ndns(self.ndns_idx['univ_ipo'], ndns_sorted=False)
        return result
    
    @context_data
    def get_univ_ipo(self):
        '''获取ndns轴， 添加listed'''
        df = self.wind_des.copy()
        df.rename(columns={'list_dt': 'indate', 'delist_dt': 'outdate'}, inplace=True)
        ndns_df = self.get_univ(df[['stockcode', 'indate', 'outdate']], right_close=False)
        result = pd.DataFrame({'univ_ipo': ndns_df.unstack(level=1)})
        return result

    @cached_property
    def wind_cal(self):
        return self.ds('asharecalendar').get_df(index=False)

    @cached_property
    def wind_des(self):
        return self.ds('asharedescription').get_df(index=False)

    def get_instruments(self, start_dt, end_dt):
        ''' 区间内上市的股票 '''
        df = self.wind_des
        return df[(df.delist_dt >= start_dt) & (df.list_dt <= end_dt)].stockcode.values

    def init_from_wind(self):
        # self.db.list_tables()
        df = self.wind_cal
        self.cn_tds = df.loc[df.exch == 'SSE'].sort_values('trade_dt')['trade_dt'].values
        self.callib = QtCalendar(self.wind_cal, all_dts=self.cn_tds)
        self.trade_dts = self.callib.get_trade_dts(self.start_dt, self.end_dt)
        self.instruments = self.get_instruments(self.start_dt, self.end_dt)
        end = self.end_dt + 30000
        rps = pd.date_range(str(self.start_dt), str(end), freq='Q-DEC')  # 外推3年 用于存预期数据
        rys = pd.date_range(str(self.start_dt), str(end), freq='Y')
        self.all_dts = pd.date_range('19900101', str(self.end_dt)).strftime("%Y%m%d").astype(int).values  # v1.4
        self.report_periods = rps.strftime("%Y%m%d").astype(int).values
        self.report_years = rys.strftime("%Y").astype(int).values
        self.ann_dts = pd.date_range(str(self.start_dt), str(self.end_dt)).strftime("%Y%m%d").astype(int).values
        self.nd, self.ns, self.nq, self.na = len(self.trade_dts), len(self.instruments), len(self.report_periods), len(
            self.ann_dts)
        self.nd_dict = pd.Series(np.arange(self.nd), self.trade_dts)
        self.ns_dict = pd.Series(np.arange(self.ns), self.instruments)
        self.nq_dict = pd.Series(np.arange(self.nq), self.report_periods)
        self.na_dict = pd.Series(np.arange(self.na), self.ann_dts)
        self.na_dict1 = pd.Series(np.arange(len(self.all_dts)), self.all_dts)  # v1.4

    def is_tds(self, dts: np.ndarray, exclude=(0, 99999999)):
        '''是否日期序列均为交易日'''
        exclude_mask = np.isin(dts, exclude)
        return np.isin(dts[~exclude_mask], self.cn_tds).all()

    def cal2td(self, dts):
        ''' 日历日期转换为trade dt '''
        return self.trade_dts[self.dt_loc(dts)]

    def isin_instrument(self, stklist):
        ''' 是否在股票池中 '''
        return np.isin(stklist, self.instruments)

    def pivot_dup(self, df, aggfunc='sum'):
        ''' pivot: with duplicates '''
        df['trade_dt'] = self.cal2td(df['trade_dt'])
        name = df.columns.drop(['stockcode', 'trade_dt'])[0]
        ndns_df = df.pivot_table(index='trade_dt', columns='stockcode', values=name, aggfunc=aggfunc) \
            .reindex(index=self.trade_dts, columns=self.instruments)
        return ndns_df

    def shift_tds(self, dts, n):
        '''把日期变成，前后n个交易日， n不能为0
        n = -1： 向前找最近的一个交易日，
        n = 0： 向后找最近的一个交易日
        如果不在交易日为首日 越界返回0和99999999
        np.searchsorted会找到最近日期靠后的一天的位置
        '''
        dts = dts.values
        tds = self.cn_tds.copy()
        loc = np.searchsorted(tds, dts) + n
        overupper = loc > len(tds) - 1
        belowlower = loc < 0
        normal = ~(overupper | belowlower)

        result = np.ones_like(dts)
        result[normal] = tds[loc[normal]]
        result[overupper] = 99999999
        result[belowlower] = 0
        return result

    def get_univ(self, df, right_close=True):
        ''' df中包含stockcode, indate, outdate; 日期nan的位置填充9999999
        right_close: outdate是否是右闭的
        '''
        assert (np.isin(['stockcode', 'indate', 'outdate'], df.columns).all())
        if not right_close:
            df['outdate'] = self.shift_tds(df['outdate'], -1)
            assert (self.is_tds(df['outdate']))
        else:
            assert (self.is_tds(df['outdate']))
        df.indate = df.indate.clip(lower=self.trade_dts[0])
        df.outdate = df.outdate.clip(upper=self.trade_dts[-1])
        df = df[df['outdate'] >= df['indate']]
        df = df[self.isin_instrument(df['stockcode'])]  # 退市后依旧在

        indate_df = df[['stockcode', 'indate']].rename(columns={'indate': 'trade_dt'})
        indate_df['value'] = 1.0
        indate_mat = self.pivot_dup(indate_df)
        outdate_df = df[['stockcode', 'outdate']].rename(columns={'outdate': 'trade_dt'})
        outdate_df['value'] = 1.0
        outdate_mat = self.pivot_dup(outdate_df)
        indate_mat.fillna(0, inplace=True)
        outdate_mat.fillna(0, inplace=True)

        result = indate_mat.cumsum() - outdate_mat.cumsum()
        result += outdate_mat

        result.replace(0, np.nan, inplace=True)
        result[result > 0] = 1.0
        return result

    def dt_loc(self, dts):
        ''' 获取 dts 在 trade_dts 上的位置，如果找不到取后值位置
        nan的位置变nan
        '''
        loc = np.searchsorted(self.trade_dts, dts)
        return np.clip(loc, a_min=0, a_max=self.nd - 1)

    def stk_loc(self, stks):
        '''返回instrument上门的排序'''
        return self.ns_dict[stks].values

    def get_ind(self, df, on):
        ''' 生成ndns的行业分类要求字段：['ind', 'stockcode', 'indate', 'outdate'];都是右闭 '''
        assert (np.isin(['stockcode', 'indate', 'outdate'], df.columns).all())
        df = df[['stockcode', 'indate', 'outdate', on]].rename(columns={on: 'ind'})

        df.indate = df.indate.clip(lower=self.trade_dts[0])
        df.outdate = df.outdate.clip(upper=self.trade_dts[-1])
        df = df[df['outdate'] >= df['indate']]
        df = df[self.isin_instrument(df['stockcode'])]
        inds = np.unique(df['ind'])
        indmap = pd.Series(np.arange(len(inds)), index=inds)

        df['indate_loc'] = self.dt_loc(df['indate'])
        df['outdate_loc'] = self.dt_loc(df['outdate'])
        df['stk_loc'] = self.stk_loc(df['stockcode'])
        df['weight'] = df['ind'].replace(indmap)

        result = np.empty([self.nd, self.ns]) * np.nan
        for i_row, row in df.iterrows():
            mat_i = np.arange(row['indate_loc'], row['outdate_loc'] + 1)  # 右闭
            mat_j = [int(row['stk_loc'])]
            idx = np.ix_(mat_i, mat_j)
            result[idx] = row['weight']
        ndns_df = pd.DataFrame(result, index=self.trade_dts, columns=self.instruments)
        return ndns_df

    def to_ndns(self, ds, ndns_sorted=True):
        ''' ds: with ndnp_idx 变成ndns_df
        ndns_sorted: False: 代表ndns有序
        '''
        # name = ds.name
        # assert (isinstance(ds, pd.Series))
        if ndns_sorted:
            assert (ds.index.names == self.ndns_pkeys)
            v = ds.values.reshape([self.ns, self.nd]).T
            result = pd.DataFrame(v, index=self.trade_dts, columns=self.instruments)
        else:
            result = ds.unstack(level=0).reindex(index=self.trade_dts, columns=self.instruments)
        return result

    def to_nans(self, ds, nans_sorted=True):
        ''' ds: with ndnp_idx 变成ndns_df
        ndns_sorted: True: 代表nans有序
        '''
        # name = ds.name
        assert (isinstance(ds, pd.Series))
        if nans_sorted:
            assert (ds.index.names == self.nans_pkeys)
            v = ds.values.reshape([self.ns, self.na]).T
            result = pd.DataFrame(v, index=self.ann_dts, columns=self.instruments)
        else:
            result = ds.unstack(level=0).reindex(index=self.ann_dts, columns=self.instruments)
        return result

    # @retry(max_attempts=5, delay=1)
    def to_feature(self, ds, feature_name, database = None, des=''):
        """
            将df转换为非结构化数据
        """
        if database is None:
            database = self.user_db_name
        print(f'writing {feature_name}')
        idx_names = ds.index.names
        df = ds.reset_index().rename(columns={0: 'value'})
        df['feature'] = feature_name
        df.drop_duplicates(subset=idx_names, inplace=True)
        self.to_parquet(database + '.feature', df, partition_by=['feature'])

        json_file = os.path.join(self.user_repo, database, 'feature', f'feature={feature_name}', 'des.json' )
        metadata = {}
        metadata['primary_keys'] = idx_names
        metadata['des'] = des
        metadata['last_update'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        write_json(metadata, json_file)
    
    # @retry(max_attempts=5, delay=1)
    def read_feature(self, feature_name, ndns = False):
        """
            读取非结构化数据
        """
        pth = os.path.join(self.user_repo, self.user_db_name, 'feature', f'feature={feature_name}')
        if not os.path.exists(pth):
            raise Exception(f'{feature_name} not exists')
        p_path = [os.path.join(pth,_i) for _i in os.listdir(pth) if _i.endswith('.parquet')]
        json_file = os.path.join(self.user_repo, self.user_db_name, 'feature', f"feature={feature_name}", 'des.json')
        metadata = read_json(json_file)
        primary_keys = metadata['primary_keys']
        df = pd.read_parquet(p_path)
        df = df.set_index(primary_keys)
        assert(len(df.columns == 1))
        ds = df[df.columns[0]]
        if ndns:
            ds = self.to_ndns(ds, True)
        return ds

    def run(self,file):
        """run cscdata init first, then use func to init here params"""
        super().run(file)

        self.init_from_wind()
        self.ndns_idx = self.get_univ_ipo()
        
        self.univ_ipo = self.get_univ_ipo_mat()
        self.nans_ipo = self.univ_ipo.reindex(index = self.ann_dts)
        self.nans_idx = pd.DataFrame({'univ_ipo' : self.nans_ipo.unstack(level=1)})
        self.nans_idx.index = self.nans_idx.index.set_names(self.nans_pkeys)
        
        # 必须传入的配置时间轴，为取历史数据使用
        localdata.initialize(self.trade_dts)

        # users func
        self._on("_init")
